# Grid World
# written by Aleksander Sadikov, October 2007
# for AI practical classes (search topics)

from random import random

class Position:
    def __init__(self, Rows = 20, Columns = 20, ObstacleProb = 0.3):
        self.GridHeight = Rows                  # number of rows in the grid
        self.GridWidth = Columns                # number of columns in the grid
        self.ObstacleProb = ObstacleProb        # probability of a square being an obstacle
        self.Grid = self.GenerateGridWorld()
        
        # start and goal position in the grid (as a complex number)
        self.State = 2 + 2j 
        self.Goal = self.GridHeight - 5 + (self.GridWidth - 5) * 1j

    def GenerateGridWorld(self):
        grid = [int(random() < self.ObstacleProb) for n in range(self.GridHeight * self.GridWidth)]
        grid[2 * self.GridHeight + 2] = 0
        grid[(self.GridHeight - 5) * self.GridWidth + self.GridWidth - 5] = 0
        return grid

    def SetStandardPosition(self):
        # sets a standard gridworld position for use in classroom
        self.GridHeight = 10
        self.GridWidth = 10
        self.ObstacleProb = 0.3
        self.Grid = [ 0,0,0,0,0,0,0,1,1,1,
                      1,1,1,0,1,0,0,1,0,0,
                      1,0,0,0,1,0,0,1,0,1,
                      0,0,1,0,0,0,1,1,1,1,
                      1,0,0,0,0,1,1,0,0,1,
                      1,1,0,1,1,0,0,1,0,0,
                      1,0,0,0,1,0,0,0,0,0,
                      1,0,0,0,1,0,0,0,0,0,
                      0,0,0,0,0,0,0,0,0,0,
                      0,0,0,0,1,1,0,0,0,0 ]
        self.State = 2 + 2j
        self.Goal = self.GridHeight - 5 + (self.GridWidth - 5) * 1j
        
    def ID(self):
        # ID is simply the same as State (single complex number)
        return self.State

    def PrintPosition(self):
        for c in range(self.GridWidth + 2): print '-',
        print
        for r in range(self.GridHeight):
            print '|',
            for c in range(self.GridWidth):
                if r + c * 1j == self.Goal: print 'O',
                elif r + c * 1j == self.State: print 'X',
                elif self.Grid[r * self.GridWidth + c]: print '#',
                else: print ' ',
            print '|'
        for c in range(self.GridWidth + 2): print '-',
        print

    def ExecuteSequence(self, sequence):
        for m in sequence:
            self.Move(m)

    def UndoSequence(self, sequence):
        for m in reversed(sequence):
            self.UndoMove(m)
    
    def Move(self, (move, cost)):
        self.State += move

    def UndoMove(self, (move, cost)):
        # undoing a move is the same as doing a move in the inverse direction
        self.State -= move      # the same as adding (move * -1) which is the inverse move

    def GenerateMoves(self):
        # simply check for boundaries and obstacles
        # the cost of all moves is 1 (diagonals could be 1.4 if they were legal)
        moves = []
        if self.State.real > 0 and not self.Grid[int((self.State.real - 1) * self.GridWidth + self.State.imag)]:
            moves.append( (-1, 1) )
        if self.State.real + 1 < self.GridHeight and not self.Grid[int((self.State.real + 1) * self.GridWidth + self.State.imag)]:
            moves.append( (1, 1) )
        if self.State.imag > 0 and not self.Grid[int(self.State.real * self.GridWidth + self.State.imag - 1)]:
            moves.append( (-1j, 1) )
        if self.State.imag + 1 < self.GridWidth and not self.Grid[int(self.State.real * self.GridWidth + self.State.imag + 1)]:
            moves.append( (1j, 1) )
        return moves

    def Solved(self):
        return self.State == self.Goal

    def Evaluate(self):
        # score can be the Manhattan distance between current state and goal state
        dist = self.Goal - self.State
        return int(abs(dist.real) + abs(dist.imag))
    
